import sys
import loguru

HANDLER_ID = 0  # The preset logging handler id
_LOG = loguru.logger


def set_log_level(log_level="ERROR"):
    global HANDLER_ID, _LOG

    # This is done to keep the formatting close to the original loguru.logger.
    # While not showcasing the <INFO> level to the user.
    _LOG.remove(HANDLER_ID)

    HANDLER_ID = _LOG.add(
        sys.stderr,
        format=(
            "<green>{time}</green> | <blue>{name}:{function}:{line}</blue> | {message}"
        ),
        colorize=True,
        level=log_level,
    )



"""
Data redaction module, will apply redaction on the values of attribute keys by the given reg rules.
"""
import json
import re
from json.decoder import JSONDecodeError

from opentelemetry.util.types import Attributes

from postman_sdk.model.data_redaction import (
    DataRedactionConfig,
)
from postman_sdk.util.logger import _LOG

# pylint: disable=W0212,W0612,import-outside-toplevel


class DataRedactor:
    """
    DataRedactor, must be loaded with Span exporter and invoke at end_span hook.
    """

    regex_redaction = {}

    def __init__(self, redaction_rules: DataRedactionConfig):
        self.__compile_rules(redaction_rules.rules)

    def __compile_rules(self, rules):
        """
        Method to parse rules and index
        :param rules:
        :return:
        """
        # Preventing circular import.
        from postman_sdk.plugins import DEFAULT_REDACTION_RULES

        combined_rules = {}

        if hasattr(rules, "__root__"):
            # In case of conflict, the items from rules will override the DEFAULT_REDACTION_RULES.
            combined_rules.update(rules.__root__)

        combined_rules.update(DEFAULT_REDACTION_RULES)
        # Making sure that user rules are given priority.

        for rl_name, regex_rule_compiled in combined_rules.items():
            self.regex_redaction[rl_name] = re.compile(
                regex_rule_compiled, re.IGNORECASE
            )

    def run(self, span_attributes: Attributes):
        """
        Method to run the redaction.
        :param span_attributes:
        :return: span_attributes
        """
        redacted_result_by_section = {}

        for req_type in ["request", "response"]:
            span_attributes, redacted_result = self.__redact(span_attributes, req_type)

            if redacted_result:
                redacted_result_by_section[req_type] = redacted_result

        if redacted_result_by_section:
            from postman_sdk.plugins import POSTMAN_DATA_REDACTION_FLAG_ATTRIBUTE_NAME

            span_attributes[POSTMAN_DATA_REDACTION_FLAG_ATTRIBUTE_NAME] = True
            # TO-DO capture what rules applied
            # span_attributes[POSTMAN_REDACTED_ATTRIBUTE_NAME] = json.dumps(redacted_result_by_section)

        return span_attributes

    def __redact(self, span_attributes, req_type):
        """
        Method to run redaction.
        :param span_attributes:
        :param req_type:
        :return:
        """
        redact_engine_map = {}

        if req_type == "request":
            from postman_sdk.plugins import HTTP_ATTRIBUTES__REQUEST_REDACTION_MAP

            redact_engine_map = HTTP_ATTRIBUTES__REQUEST_REDACTION_MAP
        elif req_type == "response":
            from postman_sdk.plugins import HTTP_ATTRIBUTES__RESPONSE_REDACTION_MAP

            redact_engine_map = HTTP_ATTRIBUTES__RESPONSE_REDACTION_MAP
        else:
            _LOG.warning("Unknown redaction request section")

            return span_attributes
        rules_applied = {}
        for section, engine_map in redact_engine_map.items():
            attribute = engine_map["attribute_key"]
            data = span_attributes.get(attribute)

            if not data:
                _LOG.debug(
                    f"No data for {attribute} for {req_type} section of the request, no"
                    " redaction applied"
                )

                continue

            _LOG.debug(f"Processing redaction {section} section on {req_type} ")

            redaction_method = engine_map["redaction_function"]

            for reg_ex_label, reg_rl in self.regex_redaction.items():
                replaced = getattr(self, redaction_method)(data, reg_rl)

                if data == replaced:
                    continue

                redacted_attributes = rules_applied.setdefault(attribute, [])

                if reg_ex_label not in redacted_attributes:
                    redacted_attributes.append(reg_ex_label)

                span_attributes[attribute] = replaced
                data = replaced

        return span_attributes, rules_applied

    @staticmethod
    def __obfuscate_json_string(json_string: str, regex_compiled):
        """
        RApply given reg ex rule on the value of the json pared object and return stringified json.
        :param regex_compiled:
        :param json_string:
        :return:
        """
        from postman_sdk.plugins import DEFAULT_REDACTION_REPLACEMENT_STRING

        try:
            json_obj = json.loads(json_string)
        except JSONDecodeError:
            _LOG.warning(f"Unable to decode incoming JSON string: {json_string}")

            return json_string

        if isinstance(json_obj, dict):
            for key_name, val in json_obj.items():
                data_val = val if isinstance(val, str) else json.dumps(val)
                data_val = regex_compiled.sub(
                    DEFAULT_REDACTION_REPLACEMENT_STRING, data_val
                )

                json_obj[key_name] = data_val
        elif isinstance(json_obj, list):
            for idx, item in enumerate(json_obj):
                data_val = item if isinstance(item, str) else json.dumps(item)

                json_obj[idx] = DataRedactor.__obfuscate_string(
                    data_val, regex_compiled
                )

        return json.dumps(json_obj)

    @staticmethod
    def __obfuscate_string(text_content, regex_compiled):
        """
        Apply given rex ex on the input string and return results.
        :param regex_compiled:
        :param text_content:
        :return:
        """
        from postman_sdk.plugins import DEFAULT_REDACTION_REPLACEMENT_STRING

        return regex_compiled.sub(DEFAULT_REDACTION_REPLACEMENT_STRING, text_content)

    @staticmethod
    def redact_headers_data(data, reg_ex_compiled):
        """
        Method to call redaction on  headers
        :param reg_ex_compiled:
        :param data:
        :return:
        """
        if not reg_ex_compiled or not data:
            return data

        return DataRedactor.__obfuscate_json_string(data, reg_ex_compiled)

    @staticmethod
    def redact_body_data(data, reg_ex_compiled):
        """
        Method to call redaction on Body
        :param reg_ex_compiled:
        :param data:
        :return:
        """
        if not reg_ex_compiled or not data:
            return data

        return DataRedactor.__obfuscate_string(data, reg_ex_compiled)

    @staticmethod
    def redact_query_data(data, reg_ex_compiled):
        """
        method to call redaction on parsed Query in dict format
        :param data:
        :param reg_ex_compiled:
        :return:
        """
        if not reg_ex_compiled or not data:
            return data

        return DataRedactor.__obfuscate_json_string(data, reg_ex_compiled)

    @staticmethod
    def redact_uristring_data(data, reg_ex_compiled):
        """
        Method to call redaction on string forma of query params.
        :param data:
        :param reg_ex_compiled:
        :return:
        """
        if not reg_ex_compiled or not data:
            return data

        return DataRedactor.__obfuscate_string(data, reg_ex_compiled)



"""
This module contains utility functions used in the Postman SDK.

"""
import json
from typing import Optional
from urllib.parse import urlparse

from postman_sdk.model.base_config import BaseConfig
from postman_sdk.receiver_service import (
    BOOTSTRAP_PATH,
    TRACE_RECEIVER_PATH,
    HEALTHCHECK_PATH,
)
from postman_sdk.util.logger import _LOG


def unpack_url(url: str):
    """
    This function takes a URL as input and returns a dictionary with two keys - "target" and "query". The "target"
    key contains the path of the URL, and the "query" key contains a JSON string with the query parameters of the URL.
    :param url: string
    :return dict
    """
    parsed_url = urlparse(url)

    return {"target": parsed_url.path, "query": parse_query_string(parsed_url.query)}


def parse_query_string(query: str) -> str:
    """
    Parses the Query string.
    :param query:
    """
    if isinstance(query, str) and query:
        query_items = query.split("&")
        query_obj = {}

        for qry in query_items:
            if "=" in qry:
                try:
                    key, val = qry.split("=")
                    query_obj[key] = val
                except:
                    _LOG.warning(f"Postman: url query parse error {qry}")

        return json.dumps(query_obj)

    return query


def add_postman_excluded_urls(config: BaseConfig):
    """
    This function returns a comma-separated string of URLs that should be excluded
    from tracing in the Postman SDK. The URLs that are excluded are the bootstrap path, trace receiver path,
    health check path, and any additional URLs specified in the Postman SDK configuration.

    **This is only applicable on outgoing requests / requests originating from the server.**
    :parm config;BaseConfig
    :return str.
    """
    excludes = [
        f"{config.receiver_base_url}{BOOTSTRAP_PATH}",
        f"{config.receiver_base_url}{TRACE_RECEIVER_PATH}",
        f"{config.receiver_base_url}{HEALTHCHECK_PATH}",
    ]

    if config.ignore_outgoing_requests:
        excludes.extend(config.ignore_outgoing_requests)

    return ",".join(excludes)


def add_user_excluded_urls(url_regexes: Optional[list]) -> Optional[str]:
    """
    This function returns a comma-separated string of URLs that should be excluded
    from tracing in the Postman SDK.

    It takes into account the user-defined exclusion list and returns a compatible version for
    downstream instrumentations.

    **This is only applicable on incoming requests to the server.**
    """
    if url_regexes:
        return ",".join(url_regexes)



"""
Module for Flask Instrumentation, will relay on open-telemetry.instrumentation.flask
"""
import json

# pylint:disable=W0718,W0613,C0123
from typing import Dict, Optional

import wrapt
from opentelemetry.sdk.trace import Span
from opentelemetry.trace import get_current_span
from postman_sdk.instrumentor.base_instrumentation import (
    HttpInstrumentationBase,
    BaseResponse,
    BaseRequest,
)
from postman_sdk.util.logger import _LOG

REQUEST_ATTRIBUTE = "request_object"


class FlaskInstrumentation(HttpInstrumentationBase):
    """
    Class for FlaskInstrumentation
    """

    FLASK_REQUEST_VAR_NAME = "werkzeug.request"

    def __init__(
        self,
        tracer_provider,
        ignore_incoming_requests: Optional[str] = None,
        *args,
        **kwargs,
    ):
        self.instrumented_instance = None
        self.tracer_provider = tracer_provider
        self.ignore_incoming_requests = ignore_incoming_requests
        module = "FlaskInstrumentor"

        try:
            super().__init__(
                opentelemetry_package="opentelemetry.instrumentation.flask",
                module_name=module,
            )

            wrapt.wrap_function_wrapper(
                "flask", "Flask.__init__", self._patch_instance_for_instrumentation
            )
        except ModuleNotFoundError:
            _LOG.warning(f"Postman: skipping instrumenting {module}")

    @staticmethod
    def request_hook(span: Span, flask_request_environ: Dict) -> None:
        """
        Request hook for flask
        :param span:
        :param flask_request_environ:
        :return:
        """
        if span is None:
            return

        try:
            request = flask_request_environ.get(
                FlaskInstrumentation.FLASK_REQUEST_VAR_NAME
            )
            request_object = BaseRequest()

            try:
                request_object.body = request.json
            except:
                request_object.body = {}

            request_object.body = json.dumps(request_object.body)
            request_object.headers = json.dumps(dict(request.headers))
            request_object.params = json.dumps(request.view_args)
            request_object.query = json.dumps(request.args)

            FlaskInstrumentation.set_attributes(span, request_object)

        except Exception as error:  # TODO fine tune error catching.
            _LOG.debug("Flask request instrumentation error: %s", error)

    @staticmethod
    def flask_response_callback(response):
        """
        Callback for flask to call before the request.
        :param response:
        :return:
        """
        span = get_current_span()
        if not span:
            _LOG.debug("Failed to get the current span.")
            return None

        try:
            response_object = BaseResponse()
            response_object.body = (
                json.dumps(response.json) if response.is_json else "{}"
            )
            response_object.headers = json.dumps(dict(response.headers))

            FlaskInstrumentation.set_attributes(span=span, http_object=response_object)
        except Exception as error:
            _LOG.debug(f"Flask after_request instrumentation error: {error}")

        return response

    def _patch_instance_for_instrumentation(self, wrapped, instance, args, kwargs):
        res = wrapped(*args, **kwargs)

        # wrapping.
        try:
            if not self.instrumented_instance:
                self.instrumented_instance = self.instrumentor_class().instrument_app(
                    instance,
                    tracer_provider=self.tracer_provider,
                    request_hook=self.request_hook,
                    excluded_urls=self.ignore_incoming_requests,
                )

                instance.after_request(self.flask_response_callback)

        except Exception as error:
            _LOG.debug(f"Flask __init__ instrumentation error: {error}.")

        return res



"""
Data tool provided redaction and truncation.
"""
from postman_sdk.plugins.data_redaction import DataRedactor
from postman_sdk.plugins.data_truncation import (
    DataTruncation,
    DEFAULT_TRUNCATION_MAX_DEPTH,
)

HTTP_REQUEST_KEYS = {"uri": "http.target", "method": "http.method"}

DEFAULT_REDACTION_REPLACEMENT_STRING = "************"

DEFAULT_REDACTION_RULES = {
    "pm_postman_api_key": r"PMAK-[a-f0-9]{24}-[a-f0-9]{34}",
    "pm_postman_access_key": r"PMAT-[0-9a-z]{26}",
    "pm_basic_auth": r"Basic (?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{4}|[A-Za-z0-9+\/]{3}=|[A-Za-z0-9+\/]{2}={2})$",
    "pm_bearer_token": r"Bearer [a-z0-9A-Z\-\._~\+\/]{15,1000}",
}

HTTP_ATTRIBUTES__REQUEST_REDACTION_MAP = {
    "body": {
        "attribute_key": "http.request.body",
        "redaction_function": "redact_body_data",
    },
    "headers": {
        "attribute_key": "http.request.headers",
        "redaction_function": "redact_headers_data",
    },
    "queryUrl": {
        "attribute_key": "http.url",
        "redaction_function": "redact_uristring_data",
    },
    "queryString": {
        "attribute_key": "http.request.query",
        "redaction_function": "redact_query_data",
    },
    "targetUrl": {
        "attribute_key": "http.target",
        "redaction_function": "redact_uristring_data",
    },
}
HTTP_ATTRIBUTES__RESPONSE_REDACTION_MAP = {
    "body": {
        "attribute_key": "http.response.body",
        "redaction_function": "redact_body_data",
    },
    "headers": {
        "attribute_key": "http.response.headers",
        "redaction_function": "redact_headers_data",
    },
}

POSTMAN_DATA_REDACTION_FLAG_ATTRIBUTE_NAME = "postman.dataRedaction"
POSTMAN_REDACTED_ATTRIBUTE_NAME = "postman.redacted"
POSTMAN_DATA_TRUNCATION_ATTRIBUTE_NAME = "postman.dataTruncated"

SPAN_HTTP_BODY_ATTRIBUTES_NAMES = {
    "response": "http.response.body",
    "request": "http.request.body",
}

DATA_REDACTION_PLUGIN_NAMESPACE = "redact_sensitive_data"
DATA_TRUNCATION_PLUGIN_NAMESPACE = "truncate_data"

PLUGINS = {
    DATA_REDACTION_PLUGIN_NAMESPACE: {
        "function": DataRedactor,
        "config_attributes": [
            {
                "name": "redact_sensitive_data",
                "function_arg": "redaction_rules",
                "default": {},
            }
        ],
    },
    DATA_TRUNCATION_PLUGIN_NAMESPACE: {
        "function": DataTruncation,
        "config_attributes": [
            {
                "name": "truncation_max_depth",
                "function_arg": "max_depth",
                "default": DEFAULT_TRUNCATION_MAX_DEPTH,
            }
        ],
    },
}



"""
Module to customize SpanExporter
"""
import json
from collections import OrderedDict
from google.protobuf.json_format import MessageToDict
from opentelemetry.sdk.trace import ReadableSpan
from opentelemetry.exporter.otlp.proto.http.trace_exporter import OTLPSpanExporter
from opentelemetry.proto.collector.trace.v1.trace_service_pb2 import (
    ExportTraceServiceRequest,
)
from postman_sdk.model.base_config import BaseConfig
from postman_sdk.model.http import EmptyResponse
from postman_sdk.plugins import (
    DATA_REDACTION_PLUGIN_NAMESPACE,
    DATA_TRUNCATION_PLUGIN_NAMESPACE,
    POSTMAN_DATA_TRUNCATION_ATTRIBUTE_NAME,
    POSTMAN_DATA_REDACTION_FLAG_ATTRIBUTE_NAME,
)

from postman_sdk.util.logger import _LOG

_OTL_JSON_CONTENT_HEADER = {"Content-Type": "application/json"}


class PostmanSpanExporter(OTLPSpanExporter):
    """
    Implementation of SpanExporter.
    """

    def __init__(self, *args, **kwargs):
        """
        exporter init.
        """
        self.postman_tracer = kwargs.pop("postman_tracer")
        self.config = kwargs.pop("config")
        self.plugins = OrderedDict()

        if self.config.truncate_data:
            self.load_plugin(DATA_TRUNCATION_PLUGIN_NAMESPACE, self.config)

        redaction_config = self.config.redact_sensitive_data

        if redaction_config and redaction_config.enable:
            self.load_plugin(DATA_REDACTION_PLUGIN_NAMESPACE, self.config)

        super().__init__(*args, **kwargs)

    def load_plugin(self, plugin_name: str, config: BaseConfig):
        """
        Method to load span hooks
        :param plugins:
        :return:
        """

        from postman_sdk.plugins import PLUGINS

        plugin = PLUGINS.get(plugin_name)

        if not plugin:
            _LOG.error(
                f"Invalid plugin: {plugin_name}, should be one of"
                f" {', '.join(PLUGINS.keys())}"
            )

            return

        func_config = {}

        for config_attrs in plugin["config_attributes"]:
            func_config[config_attrs["function_arg"]] = getattr(
                config,
                config_attrs["name"],
                config_attrs["default"],
            )

        self.plugins[plugin_name] = plugin["function"](**func_config)

    def run_plugins(self, spans) -> list:
        """
        Apply plugins on spans.
        :param spans:
        :return: list of spans.
        """
        for span_idx, span in enumerate(spans):
            for plugin_name, plugin in self.plugins.items():
                _LOG.info(f"Running plugin in span: {plugin_name}")

                try:
                    """
                    Structure of span.get("attributes") is:
                    [
                        {
                            "key": "http.request.body",
                            "value": {"stringValue": "<json-dumped-body-here>"},
                        }
                    ]
                    """
                    span_attributes = span.get("attributes", [])

                    """
                    parsed_attributes = {
                        "http.request.body": {"stringValue": "<body>"}
                    }
                    """
                    parsed_attributes = {
                        item["key"]: item["value"] for item in span_attributes
                    }
                    """
                    plugin_attributes = {
                        "http.request.body": "<body>"
                    }
                    """
                    plugin_attributes = {
                        key: value[list(value.keys())[0]]
                        for key, value in parsed_attributes.items()
                    }

                    _attributes = plugin.run(plugin_attributes)
                    if _attributes:
                        parsed_attributes = {
                            key: {list(value.keys())[0]: _attributes.get(key, "")}
                            for key, value in parsed_attributes.items()
                        }

                        span["attributes"] = [
                            {"key": key, "value": value}
                            for key, value in parsed_attributes.items()
                        ]
                        if plugin_name == DATA_TRUNCATION_PLUGIN_NAMESPACE:
                            span["attributes"].extend(
                                [
                                    {
                                        "key": POSTMAN_DATA_TRUNCATION_ATTRIBUTE_NAME,
                                        "value": {"boolValue": True},
                                    }
                                ]
                            )
                        if plugin_name == DATA_REDACTION_PLUGIN_NAMESPACE:
                            span["attributes"].extend(
                                [
                                    {
                                        "key": POSTMAN_DATA_REDACTION_FLAG_ATTRIBUTE_NAME,
                                        "value": {"boolValue": True},
                                    }
                                ]
                            )
                except Exception as plugin_error:
                    spans[span_idx] = "plugin_failed"

                    _LOG.warning(
                        f"Failed to run {plugin_name} plugin, this span will be dropped."
                    )
                    _LOG.debug(f"Failed span is - {span}")
                    _LOG.error(plugin_error)
                    # Skip other plugins if one fails.
                    break
        return spans

    def _export(self, serialized_data: str):
        """
        :param serialized_data:
        :return:
        """

        if self.postman_tracer.is_suppressed():
            return EmptyResponse()

        # Note: Explore if Multi-process, multi-thread can cause
        # session conflicts ?
        self._session.headers.update(_OTL_JSON_CONTENT_HEADER)
        export_request = ExportTraceServiceRequest.FromString(serialized_data)
        json_data = MessageToDict(export_request)

        if self.plugins:
            scope_spans = json_data.get("resourceSpans", [{}])[0].get(
                "scopeSpans", [{}]
            )
            for scope_span in scope_spans:
                spans = scope_span.get("spans", [])
                if spans:
                    modified_spans = self.run_plugins(spans=spans)

                    modified_spans = list(
                        filter(lambda span: span != "plugin_failed", modified_spans)
                    )

                    scope_span["spans"] = modified_spans

        _LOG.debug(f"Postman SDK: Trace Service Data Exported: {json_data}")
        resp = self._session.post(
            url=self._endpoint,
            data=json.dumps(json_data),
            verify=self._certificate_file,
            timeout=self._timeout,
        )

        if resp.status_code:
            _LOG.debug(f"Postman SDK Trace Service Response {resp.status_code}")

        return resp



"""
Instrumentation for Python `requests` library.
"""
import json
from typing import Optional

from postman_sdk.instrumentor.base_instrumentation import (
    HttpInstrumentationBase,
    BaseResponse,
    BaseRequest,
)
from postman_sdk.util.helper import unpack_url
from postman_sdk.util.logger import _LOG


class RequestsInstrumentation(HttpInstrumentationBase):
    """
    Requests Instrumentation module
    """

    def __init__(
        self,
        tracer_provider,
        ignore_outgoing_requests: Optional[str] = None,
        *args,
        **kwargs,
    ):
        module_name = "RequestsInstrumentor"
        try:
            super().__init__(
                opentelemetry_package="opentelemetry.instrumentation.requests",
                module_name=module_name,
            )

            self.instrumented_instance = self.instrumentor_class().instrument(
                trace_provider=tracer_provider,
                span_callback=self.span_callback,
                excluded_urls=ignore_outgoing_requests,
            )
        except ModuleNotFoundError:
            _LOG.warning(f"Postman: skipping instrumenting {module_name}")

    @staticmethod
    def span_callback(span, http_object):
        """
        call back method to requests instrumentor.
        :param span
        :param http_object
        """
        if not http_object:
            _LOG.info("Postman: Requests instrumentation empty object on callback.")

            return
        if hasattr(http_object, "request"):
            request = BaseRequest()
            body = {}

            try:
                body = http_object.request.json
            except:
                body = {}

            request.body = json.dumps(body)

            request.headers = json.dumps(dict(http_object.request.headers))
            url_object = unpack_url(http_object.request.url)
            request.route = url_object["target"]
            request.query = url_object["query"]

            RequestsInstrumentation.set_attributes(span, request)

        # Response part of the requests
        if hasattr(http_object, "raw"):
            response = BaseResponse()
            response.headers = json.dumps(dict(http_object.raw.headers))
            response.body = http_object.text

            RequestsInstrumentation.set_attributes(span, response)



"""
Provide basic config structure
"""
# pylint:disable=E0611,R0903
from typing import Optional, List

try:
    from typing_extensions import TypedDict
except ImportError:
    from typing import TypedDict

from pydantic import BaseModel

from postman_sdk.model.data_redaction import DataRedactionConfig
from postman_sdk.util.constants import DEFAULT_RECEIVER_BASE_URL


class BaseConfig(BaseModel):
    """
    Base config class
    """

    collection_id: str
    receiver_base_url: str = DEFAULT_RECEIVER_BASE_URL
    enable: Optional[bool] = True
    buffer_interval_in_milliseconds: Optional[int]
    api_key: str
    debug: Optional[bool] = False
    truncate_data: Optional[bool] = True
    redact_sensitive_data: Optional[DataRedactionConfig]
    ignore_outgoing_requests: Optional[List[str]]
    ignore_incoming_requests: Optional[List[str]]


class PostmanSDKConfig(TypedDict, total=False):
    """
    Typed Dict representation of the above BaseConfig.
    This is required to allow for a more UX friendly onboarding for the users.
    """

    collection_id: str
    api_key: str
    enable: Optional[bool]
    debug: Optional[bool]
    truncate_data: Optional[bool]
    redact_sensitive_data: Optional[DataRedactionConfig]
    ignore_incoming_requests: Optional[List[str]]
    ignore_outgoing_requests: Optional[List[str]]
    receiver_base_url: Optional[str]
    buffer_interval_in_milliseconds: Optional[int]



"""
Defaults constants
"""

HEALTH_CHECK_ERROR_COUNT_THRESHOLD = 5
HTTP_STATUS_OK = 200
# TODO: decide proper timeout values
HTTP_REQUEST_CONNECTION_TIMEOUT = 6  # must be > multiples of 3
HTTP_REQUEST_READ_TIMEOUT = 30  # must be > multiples of 3
HTTP_REQUEST_TIMEOUT = (HTTP_REQUEST_CONNECTION_TIMEOUT, HTTP_REQUEST_READ_TIMEOUT)
DEFAULT_IGNORED_HOSTNAMES = [
    "newrelic.com",
]
POSTMAN_TELEMETRY = "postman-sdk"
POSTMAN_COLLECTION_ID = "postman.collection.id"
POSTMAN_SDK_VERSION = "postman.sdk.version"
DEFAULT_RECEIVER_BASE_URL = "https://trace-receiver.postman.com"
X_API_KEY = "x-api-key"
CONTENT_TYPE = "Content-type"

# DataTruncation constants


DEFAULT_TRUNCATION_MAX_DEPTH = 2
TRUNCATION_START_LEVEL = 1
POSTMAN_SDK_ENABLE_FLAG = "POSTMAN_SDK_ENABLE"



"""
Health check module
"""
import threading

# pylint: disable=W0622
from threading import Thread
from time import sleep
from http import HTTPStatus


from postman_sdk.receiver_service.client import call_reciever_api
from postman_sdk.receiver_service.config import (
    HEALTHCHECK_PATH,
    DEFAULT_HEALTH_PING_INTERVAL_SECONDS,
    EXPONENTIAL_BACKOFF_BASE,
)
from postman_sdk.util.constants import (
    HEALTH_CHECK_ERROR_COUNT_THRESHOLD,
)
from postman_sdk.util.logger import _LOG


class HealthCheck:
    """Health check module."""

    def __init__(self, postman_tracer):
        self.postman_tracer = postman_tracer
        self.lock = threading.Lock()
        self.default_health_ping_interval = DEFAULT_HEALTH_PING_INTERVAL_SECONDS

    def run(self):
        """Run health check in thread.

        Args:
            tracer(_type_): PostmanTracer
        """
        daemon = Thread(
            target=self.__health_poller,
            args=[],
            daemon=True,  # Main thread doesn't wait for this thread before being finished
            name="Postman SDK heath check poller",
        )
        daemon.start()
        return daemon

    def call_health_check_api(self):
        payload = {
            "sdk": {
                "collectionId": f"{self.postman_tracer.config.collection_id}",
                "enabled": self.postman_tracer.config.enable,
            }
        }

        return call_reciever_api(
            config=self.postman_tracer.config, path=HEALTHCHECK_PATH, payload=payload
        )

    def __health_poller(self, retry_count=0):
        if not self.default_health_ping_interval:
            return

        while True:
            if retry_count > 1:
                _LOG.debug(f"Retrying healthcheck retry no: {retry_count}")
            resp = self.call_health_check_api()

            if resp.status_code == HTTPStatus.OK.value:
                if resp.json().get("healthy"):
                    retry_count = 0
                    self.postman_tracer.unsuppress()

                sleep(self.default_health_ping_interval)

            elif resp.status_code == HTTPStatus.CONFLICT.value:
                if not self.postman_tracer.bootstrap_sdk():
                    # This turns off the health-check thread
                    return

                sleep(self.default_health_ping_interval)

            elif resp.status_code == HTTPStatus.NOT_FOUND.value:
                if "healthy" in resp.json() and not resp.json()["healthy"]:
                    if not self.postman_tracer.bootstrap_sdk():
                        # This turns off the health-check thread
                        return
                    sleep(self.default_health_ping_interval)
                else:
                    # Case when url itself is wrong
                    self.postman_tracer.suppress()
                    # This turns off the health-check thread
                    return

            else:
                self.postman_tracer.suppress()
                if retry_count > HEALTH_CHECK_ERROR_COUNT_THRESHOLD:
                    _LOG.warning("Shutting down Postman SDK")
                    self.postman_tracer.disable()
                    # This turns off the health-check thread
                    return

                retry_count += 1
                sleep(
                    (EXPONENTIAL_BACKOFF_BASE**retry_count)
                    * self.default_health_ping_interval
                )



"""
Enum for http verbs.
"""
from enum import Enum


class HTTPVerbEnum(str, Enum):
    """
    Http verbs enum
    """

    POST = "POST"
    GET = "GET"
    PATCH = "PATCH"
    DELETE = "DELETE"
    HEAD = "HEAD"
    PUT = "PUT"

    @staticmethod
    def to_string():
        """
        enum class to string
        :return:
        """
        return ",".join([en for en in HTTPVerbEnum])


class EmptyResponse:
    """
    Util class to mock span export response.
    """

    def __init__(self, status_code=200, data=dict()) -> None:
        self.status_code = status_code
        self.data = data

    def json(self):
        return self.data



"""
Constants for Postman receiver.
"""
BOOTSTRAP_PATH = "/sdk/bootstrap"
TRACE_RECEIVER_PATH = "/traces"
HEALTHCHECK_PATH = "/sdk/health"

DEFAULT_HEALTH_PING_INTERVAL_MILLISECONDS = 60000
DEFAULT_HEALTH_PING_INTERVAL_SECONDS = DEFAULT_HEALTH_PING_INTERVAL_MILLISECONDS // 1000

EXPONENTIAL_BACKOFF_BASE = 2
BOOTSTRAP_RETRY_COUNT = 2



"""
Data truncation module
"""
import json
from json import JSONDecodeError

from opentelemetry.util.types import Attributes

from postman_sdk.util.constants import (
    DEFAULT_TRUNCATION_MAX_DEPTH,
    TRUNCATION_START_LEVEL,
)


# pylint: disable=import-outside-toplevel


class DataTruncation:
    """
    DataTruncation plugin, plugin reshapes the json payload with <key>:<type>. level will define how deeps this should
    apply.
    """

    level: int

    def __init__(self, max_depth=DEFAULT_TRUNCATION_MAX_DEPTH):
        """
        Constructor
        :param level:
        """
        self.max_depth = max_depth

    def run(self, span_attributes: Attributes):
        """
        entry point to run the redaction.
        :param span_attributes:
        :return: span_attributes
        """

        from postman_sdk.plugins import (
            SPAN_HTTP_BODY_ATTRIBUTES_NAMES,
            POSTMAN_DATA_TRUNCATION_ATTRIBUTE_NAME,
        )

        for _, span_attrib_name in SPAN_HTTP_BODY_ATTRIBUTES_NAMES.items():
            data = span_attributes.get(span_attrib_name)

            if data:
                truncated_data = self._truncate_data(data)
                span_attributes[span_attrib_name] = truncated_data

        span_attributes[POSTMAN_DATA_TRUNCATION_ATTRIBUTE_NAME] = True

        return span_attributes

    def _truncate_data(self, data, current_level=TRUNCATION_START_LEVEL):
        """
        :param data:
        :param current_level:
        :return:
        """
        string_data_input = False

        if not data:
            return data

        if isinstance(data, str):
            string_data_input = True

            try:
                data = json.loads(data)
            except JSONDecodeError:
                return data

        if not isinstance(data, dict) and not isinstance(data, list):
            return {"type": type(data).__name__}

        # We can have list or dict by this part so, will use enumerate to help iterate.
        # enumerate results in index and item in case of list and index of a key, key in case of dict
        for idx, dat_key in enumerate(data):
            dat_val = data[dat_key] if isinstance(data, dict) else data[idx]
            data_type = {"type": type(dat_val).__name__}

            if isinstance(dat_val, dict) or isinstance(dat_val, list):
                if current_level <= self.max_depth:
                    data_type = self._truncate_data(
                        data=dat_val, current_level=current_level + 1
                    )
                elif not dat_val:
                    if isinstance(dat_val, dict):
                        data[dat_key] = data_type
                    else:
                        data[idx] = data_type
                else:
                    if isinstance(data, dict):
                        data[dat_key] = {"type": type(dat_val).__name__}
                    else:
                        data[idx] = data_type

            if isinstance(data, dict):
                data[dat_key] = data_type
            else:
                data[idx] = data_type

        if string_data_input:
            data = json.dumps(data)

        return data



"""
Module for adding all custom span keys
"""


class HTTPSpanAttributes:
    """
    Http span attributed Semantic module.
    """

    HTTP_REQUEST_BODY = "http.request.body"
    HTTP_REQUEST_HEADERS = "http.request.headers"
    HTTP_REQUEST_PARAMS = "http.request.params"
    HTTP_REQUEST_QUERY = "http.request.query"
    HTTP_ROUTE = "http.route"
    HTTP_RESPONSE_HEADERS = "http.response.headers"
    HTTP_RESPONSE_BODY = "http.response.body"



"""
Base HTTP instrumentation module, this module will handle setting attributes on spans.
Initializing the passed Frameworks.
"""
from importlib import import_module
from typing import Union

from opentelemetry.sdk.trace import Span

from postman_sdk.model.semantic_attributes import HTTPSpanAttributes


class BaseRequest:
    """
    Basic Request structure for common http spans
    """

    body: str = ""
    headers: str = ""
    params: str = ""
    query: str = ""
    route: str = ""


class BaseResponse:
    """
    Basic Response structure for common http spans
    """

    body: str = ""
    headers: str = ""


class HttpInstrumentationBase:
    """
    Basic framework for HTTP based instrumentation
    """

    def __init__(self, opentelemetry_package: str, module_name: str):
        """
        HttpInstrumentationBase, takes opentelemetry_package and module name to facilitate dynamic loading.
        """
        self.instrumentor_class = None
        self.import_and_set_instrumentor_class(opentelemetry_package, module_name)

    @staticmethod
    def set_attributes(span: Span, http_object: Union[BaseRequest, BaseResponse]):
        """
        Method to set attributed on Span for Request and Response sections.
        """
        if isinstance(http_object, BaseRequest):
            span.set_attributes(
                {
                    HTTPSpanAttributes.HTTP_REQUEST_BODY: http_object.body,
                    HTTPSpanAttributes.HTTP_REQUEST_HEADERS: http_object.headers,
                    HTTPSpanAttributes.HTTP_REQUEST_PARAMS: http_object.params,
                    HTTPSpanAttributes.HTTP_REQUEST_QUERY: http_object.query,
                    # url_rule is None in case of 404
                    HTTPSpanAttributes.HTTP_ROUTE: http_object.route,
                }
            )
        elif isinstance(http_object, BaseResponse):
            span.set_attributes(
                {
                    HTTPSpanAttributes.HTTP_RESPONSE_HEADERS: http_object.headers,
                    HTTPSpanAttributes.HTTP_RESPONSE_BODY: http_object.body,
                }
            )

    def import_and_set_instrumentor_class(self, opentelemetry_package, module_name):
        """
        Method to call the instrumentor class as member of HttpInstrumentationBase.
        """
        self.instrumentor_class = getattr(
            import_module(opentelemetry_package), module_name
        )



"""
module for data redaction models.
"""
# pylint:disable=E0611,R0903
from typing import Dict, Optional

from pydantic import BaseModel

from postman_sdk.model.dynamic_model import DynamicBaseModel


class Rules(DynamicBaseModel):
    """
    Rules object
    """

    __root__: Dict[str, str]


class DataRedactionConfig(BaseModel):
    """
    DataRedaction config model.
    """

    enable: bool = True
    rules: Optional[Rules]



import backoff
from http import HTTPStatus

from postman_sdk.receiver_service import BOOTSTRAP_PATH
from postman_sdk.receiver_service.client import call_reciever_api
from postman_sdk.receiver_service.config import BOOTSTRAP_RETRY_COUNT
from postman_sdk.model.base_config import BaseConfig


def is_retryable(resp) -> bool:
    if not hasattr(resp, "status_code"):
        return False
    return resp.status_code in [
        HTTPStatus.CONFLICT.value,
        HTTPStatus.INTERNAL_SERVER_ERROR.value,
        HTTPStatus.GATEWAY_TIMEOUT.value,
        HTTPStatus.SERVICE_UNAVAILABLE.value,
        HTTPStatus.BAD_GATEWAY.value,
    ]


def is_success(resp) -> bool:
    return resp.status_code == HTTPStatus.OK.value and "currentConfig" in resp.json()


@backoff.on_predicate(
    backoff.expo,
    predicate=is_retryable,
    max_tries=BOOTSTRAP_RETRY_COUNT,
)
def bootstrapSDK(config: BaseConfig):
    """
    method to call Bootstrap endpoint. Raises error if bootstrap errors.
    :param config:
    :return: None
    """

    payload = {
        "sdk": {"collectionId": f"{config.collection_id}", "enabled": config.enable}
    }
    resp = call_reciever_api(config=config, path=BOOTSTRAP_PATH, payload=payload)

    return resp



"""
general instrumentor initialization module.
"""
from postman_sdk.instrumentor.postman_flask_instrumentation import FlaskInstrumentation
from postman_sdk.instrumentor.postman_requests_instrumentation import (
    RequestsInstrumentation,
)


def get_instrumentations(*args, **kwargs):
    """
    Initialize all possible instrumentation frameworks.
    """
    instrumentation = [FlaskInstrumentation, RequestsInstrumentation]

    return [
        instrumentation_instance(*args, **kwargs)
        for instrumentation_instance in instrumentation
    ]



"""
SDK init module
"""
from postman_sdk.model.base_config import PostmanSDKConfig
from postman_sdk.postman_otel.postman_tracer import PostmanTracer


def initialize(config: PostmanSDKConfig):
    """
    Initialize method

    collection_id: str
        The live collection id.

    api_key: str
        Postman API key for authentication.

    enable: Optional[bool] = True
        Enable or disable the SDK. Disabled SDK does not capture any new traces,
        nor does it use up system resources.
        Can be set by using an environment variable 'POSTMAN_SDK_ENABLE',
        which will over-ride any other config.

    debug:  Optional[bool] = False
        Enable debug logs.

    truncate_data: Optional[bool] = True
        Truncate the request and response body so that no PII data is sent to Postman.

        This is **enabled** by default. Disabling it sends actual request and response payloads.

        Example:

        Sample payload or non-truncated payload:
            {
                "first_name": "John",
                "age": 30
            }

        Truncated payload:
            {
                "first_name" : {
                    "type": "str"
                },
                "age": {
                    "type": "int"
                }
            }

    redact_sensitive_data: Optional[DataRedactionConfig]
        Redact sensitive data such as api_keys and auth tokens, before they leave the sdk.

        When this is enabled, below redaction rules are applied by default (they are not case-sensitive):

        {
            "pm_postman_api_key": r"PMAK-[a-f0-9]{24}-[a-f0-9]{34}",
            "pm_postman_access_key": r"PMAT-[0-9a-z]{26}",
            "pm_basic_auth": r"Basic (?:[A-Za-z0-9+\/]{4})*(?:[A-Za-z0-9+\/]{4}|[A-Za-z0-9+\/]{3}=|[A-Za-z0-9+\/]{2}={2})$",
            "pm_bearer_token": r"Bearer [a-z0-9A-Z\-\._~\+\/]{15,1000}"
        }

        Example:
            {
                "redact_sensitive_data": {
                    "enable": True (default)
                    "rules": {
                        "<rule name>": "<regex to match the rule>", such as
                        "basic_auth": r"\b[A-Za-z0-9._%+-]+@[A-Za-z0-9.-]+\.[A-Z|a-z]{2,7}\b"
                    }
                }
            }

    ignore_incoming_requests: Optional[list]
        List of regexes to be ignored from instrumentation.
        This rule only applies to endpoints that are **served** by the application/server.

        Example:
            {
                "ignore_incoming_requests": ["knockknock", "^get.*"]
            }

            Will ignore any endpoint that contains the work "knockknock" in it, and all endpoints
            that start with get, and contain any characters after that.

    ignore_outgoing_requests: Optional[list]
        List of regexes to be ignored from instrumentation.
        This rule only applies to endpoints that are **called** by the application/server.

        Example:
            {
                "ignore_outgoing_requests": ["knockknock", "^get.*"]
            }

            Will ignore any endpoint that contains the work "knockknock" in it, and all endpoints
            that start with get, and contain any characters after that.

    receiver_base_url: str = DEFAULT_RECEIVER_BASE_URL
        Where the data should be shipped to receiver's http endpoint.
        **Do not override this value**.

    buffer_interval_in_milliseconds: Optional[int] = 5000
        The interval in milliseconds that the SDK waits before sending data to Postman.
        The default interval is 5000 milliseconds. This interval can be tweaked for lower or higher
        throughput systems.
    """

    PostmanTracer(config=config)



"""
Custom Model supporting dynamic fields.
"""
# pylint:disable=E0611
from typing import Dict

from pydantic import BaseModel


class DynamicBaseModel(BaseModel):
    """
    Pydantic custom model for dynamic fields support
    """

    def get_keys(self):
        """
        Get all the key names
        :return:
        """
        if hasattr(self, "__root__") and isinstance(self.__root__, Dict):
            return self.__root__.keys()

        return None

    def get_items(self):
        """
        Get iter of fields, values
        :return:
        """
        if hasattr(self, "__root__") and isinstance(self.__root__, Dict):
            return self.__root__.items()

        return None

    def get_entries(self):
        """
        Get list items if dynamic fields are list.
        :return:
        """
        if hasattr(self, "__root__") and isinstance(self.__root__, list):
            return self.__root__

        return None

    def get(self, key):
        """
        Check if the field name exists and return value if found.
        :param key:
        :return:
        """
        if hasattr(self, "__root__") and isinstance(self.__root__, Dict):
            return self.__root__.get(key)

        return None



import json
import requests
from http import HTTPStatus

from postman_sdk.util.constants import (
    X_API_KEY,
    CONTENT_TYPE,
)
from postman_sdk.util.logger import _LOG
from postman_sdk.model.base_config import BaseConfig
from postman_sdk.model.http import EmptyResponse
from postman_sdk.util.constants import HTTP_REQUEST_TIMEOUT


def call_reciever_api(config: BaseConfig, path: str, payload: dict):
    """
    Method to call reciever service APIs
    :param config:
    :return: None
    """
    resp = None
    headers = {X_API_KEY: config.api_key, CONTENT_TYPE: "application/json"}
    api_url = f"{config.receiver_base_url}{path}"

    try:
        resp = requests.post(
            api_url,
            data=json.dumps(payload),
            headers=headers,
            timeout=HTTP_REQUEST_TIMEOUT,
        )
        _LOG.debug(
            f"Tracer service called path: {path}, status_code:"
            f" {resp.status_code}, body: {resp.json()}"
        )
    except Exception as e:
        status_code = HTTPStatus.INTERNAL_SERVER_ERROR.value
        if resp is not None and hasattr(resp, "status_code"):
            status_code = resp.status_code
        resp = EmptyResponse(status_code=status_code, data={"message": str(e)})
        _LOG.debug(
            f"Tracer service called path: {path}, status_code:{resp.status_code},"
            f" body: {resp.json()}"
        )
    return resp



"""
SDK initializer.
"""
# pylint:disable=W0622,R0913
# system imports
try:
    from importlib_metadata import version
except ImportError:
    from importlib.metadata import version
import os
from typing import Union, Dict


# open-telemetry imports
from opentelemetry import trace
from opentelemetry.instrumentation.propagators import (
    set_global_response_propagator,
    TraceResponsePropagator,
)
from opentelemetry.sdk.resources import Resource
from opentelemetry.sdk.trace import TracerProvider
from opentelemetry.sdk.trace.export import BatchSpanProcessor
from opentelemetry.semconv.resource import ResourceAttributes
from pydantic import ValidationError

from postman_sdk.instrumentor import get_instrumentations
from postman_sdk.model.base_config import BaseConfig, PostmanSDKConfig
from postman_sdk.postman_otel.postman_exporter import PostmanSpanExporter
from postman_sdk.receiver_service import HealthCheck, TRACE_RECEIVER_PATH
from postman_sdk.receiver_service import bootstrap
from postman_sdk.util.constants import (
    POSTMAN_SDK_VERSION,
    POSTMAN_COLLECTION_ID,
    POSTMAN_TELEMETRY,
    X_API_KEY,
    POSTMAN_SDK_ENABLE_FLAG,
)
from postman_sdk.util.helper import add_postman_excluded_urls, add_user_excluded_urls
from postman_sdk.util.logger import set_log_level, _LOG


class PostmanTracer:
    """
    PostmanTracer class, that initializes otel instrumentation.
    """

    __provider: TracerProvider
    config: BaseConfig
    tracing_suppressed = False

    def __init__(self, config: Union[Dict, PostmanSDKConfig]):
        """
        initialize
        * Tracer
        * Adds SpanProcessor
        """
        try:
            self.__parse_config(config)
        except ValidationError as config_error:
            _LOG.error(f"Unable to parse configuration \n error: {config_error}")
            _LOG.warning("Postman SDK Disbaled")

            return

        set_log_level(log_level="DEBUG" if self.config.debug else "ERROR")

        POSTMAN_SDK_ENABLE = os.environ.get(POSTMAN_SDK_ENABLE_FLAG, "")

        if POSTMAN_SDK_ENABLE.lower() == "true":
            self.config.enable = True
        elif POSTMAN_SDK_ENABLE.lower() == "false":
            self.config.enable = False

        if not self.config.enable:
            _LOG.error("Postman SDK Disabled")

            return

        resource = {}

        if not self.bootstrap_sdk():
            return

        HealthCheck(self).run()

        self.__update_resource_tags(resource, self.config)

        set_global_response_propagator(TraceResponsePropagator())
        self.__provider = TracerProvider(resource=Resource.create(resource))
        trace.set_tracer_provider(self.__provider)

        exporter = self.get_otlp_exporter(self.config)

        self.__provider.add_span_processor(
            BatchSpanProcessor(
                span_exporter=exporter,
                schedule_delay_millis=self.config.buffer_interval_in_milliseconds,
            )
        )

        get_instrumentations(
            tracer_provider=self.__provider,
            ignore_outgoing_requests=add_postman_excluded_urls(self.config),
            ignore_incoming_requests=add_user_excluded_urls(
                self.config.ignore_incoming_requests
            ),
        )

        _LOG.error("Postman SDK Instrumentation successful !!")

    def bootstrap_sdk(self) -> bool:
        response = bootstrap.bootstrapSDK(self.config)
        if not bootstrap.is_success(response):
            _LOG.error("Postman SDK disabled")
            self.disable()

            return False

        self.config.enable = (
            response.json().get("currentConfig", {}).get("enabled", False)
        )

        if not self.config.enable:
            self.suppress()
        else:
            self.unsuppress()

        return True

    @staticmethod
    def get_otlp_exporter(config: BaseConfig) -> PostmanSpanExporter:
        """
        Create Span exporter and return.
        :param config: BaseConfig
        :return: PostmanSpanExporter object.
        """
        endpoint = f"{config.receiver_base_url}{TRACE_RECEIVER_PATH}"
        headers = {
            X_API_KEY: config.api_key,
        }

        return PostmanSpanExporter(
            endpoint=endpoint,
            headers=headers,
            config=config,
            postman_tracer=PostmanTracer,
        )

    @staticmethod
    def __update_resource_tags(resource_tags, config: BaseConfig):
        """
        Method to add resource tags.
        :return:
        """
        resource_tags.update(
            {
                ResourceAttributes.TELEMETRY_SDK_NAME: POSTMAN_TELEMETRY,
                POSTMAN_COLLECTION_ID: config.collection_id,
                POSTMAN_SDK_VERSION: version("postman-sdk"),
            }
        )

    def __parse_config(self, config):
        """
        Method to parse config to BaseConfig
        :param config:
        :return:
        """
        self.config = BaseConfig(**config)

        if self.config.receiver_base_url.endswith("/"):
            self.config.receiver_base_url = self.config.receiver_base_url[:-1]

    @staticmethod
    def suppress():
        _LOG.debug("Postman SDK Disabling tracing")
        PostmanTracer.tracing_suppressed = True

    @staticmethod
    def unsuppress():
        _LOG.debug("Postman SDK Enabling Tracing")
        PostmanTracer.tracing_suppressed = False

    @staticmethod
    def is_suppressed():
        return PostmanTracer.tracing_suppressed

    @staticmethod
    def disable():
        PostmanTracer.suppress()
        # TODO: Is this needed ?
        # since we've solved disabling from HealthCheck._health_poller()
        # PostmanTracer.disable_health_check()
